package com.chilunyc.fumao.config.web.security;

import com.chilunyc.fumao.common.context.Request;
import com.chilunyc.fumao.common.context.RequestHolder;
import com.chilunyc.fumao.common.exception.SystemException;
import com.chilunyc.fumao.common.response.JsonResponse;
import com.chilunyc.fumao.common.shiro.AuthUser;
import com.chilunyc.fumao.common.util.jwt.JwtUtils;
import com.chilunyc.fumao.config.web.arguments.LoginUserArgumentResolver;
import io.jsonwebtoken.Claims;
import lombok.extern.log4j.Log4j;
import org.apache.commons.lang.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.MessageSource;
import org.springframework.context.NoSuchMessageException;
import org.springframework.security.authentication.UsernamePasswordAuthenticationToken;
import org.springframework.security.core.authority.SimpleGrantedAuthority;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.stereotype.Component;
import org.springframework.web.filter.OncePerRequestFilter;
import org.springframework.web.util.ContentCachingRequestWrapper;

import javax.servlet.FilterChain;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.util.Collections;
import java.util.List;
import java.util.Locale;
import java.util.stream.Collectors;

/**
 * 把JWT转换为security的对象
 * Created by Tkk on 2018/7/23.
 */
@Component
@Log4j
public class SecurityFilter extends OncePerRequestFilter {

    @Autowired
    MessageSource messageSource;

    @Override
    protected void doFilterInternal(HttpServletRequest httpServletRequest, HttpServletResponse httpServletResponse, FilterChain filterChain) throws ServletException, IOException {
        // 区分什么请求
        Request build = Request
                .builder()
                .request(httpServletRequest)
                .build();
        String servletPath = httpServletRequest.getServletPath();
        String prefix = org.apache.commons.lang3.StringUtils.substringBetween(servletPath, "/", "/");
        switch (prefix) {
            case "system":
                build.setJwtType(JwtUtils.T.System);
                break;
            case "company":
                build.setJwtType(JwtUtils.T.Company);
                break;
            case "place":
                build.setJwtType(JwtUtils.T.Market);
                break;
            case "customer":
                build.setJwtType(JwtUtils.T.APP_C);
                break;
            case "merchant":
                build.setJwtType(JwtUtils.T.APP_B);
                break;
        }

        try {
            RequestHolder.set(build);
            String token = httpServletRequest.getHeader("token");
            if (StringUtils.isNotBlank(token)) {
                //
                Claims claims = JwtUtils.decode(build.getJwtType(), httpServletRequest.getHeader("token"));
                AuthUser authUser = toAuthUser(claims);

                // 放入jwt数值, 用于后面转换方法对面
                build.setJwtValue(claims);

                // 如果能转换成功, 说明登录没问题, 那么转换为security对象, 做权限处理
                UsernamePasswordAuthenticationToken usernamePasswordAuthenticationToken = toAuthToken(authUser);
                SecurityContextHolder.getContext().setAuthentication(usernamePasswordAuthenticationToken);
            }

            // 包装request对象, 可以之后做日志处理, 可以反复读取请求体
            ContentCachingRequestWrapper r = new ContentCachingRequestWrapper(httpServletRequest);
            filterChain.doFilter(r, httpServletResponse);
        }
        // 业务一次
        catch (SystemException e) {
            log.error(e.getMessage(), e);
            httpServletResponse.setCharacterEncoding("UTF-8");
            httpServletResponse.getWriter().write(
                    JsonResponse.fail(e.getErrorCode(), getMessage(e.getErrorCode(), e.getParams())).toString()
            );
        }
        // 所有异常
        catch (Exception e) {
            log.error(e.getMessage(), e);
            httpServletResponse.setCharacterEncoding("UTF-8");
            httpServletResponse.getWriter().write(
                    JsonResponse.fail().toString()
            );
        } finally {
            RequestHolder.clean();
        }
    }

    /**
     * @param claims
     * @return
     */
    private AuthUser toAuthUser(Claims claims) {
        List<String> roles = (List<String>) claims.get("role");
        if (roles == null) {
            roles = Collections.emptyList();
        }
        return AuthUser
                .builder()
                .id(Long.parseLong(claims.getSubject()))
                .authorities(
                        roles.stream()
                                .map(SimpleGrantedAuthority::new)
                                .collect(Collectors.toList()))
                .build();
    }

    private UsernamePasswordAuthenticationToken toAuthToken(AuthUser authUser) {
        return new UsernamePasswordAuthenticationToken(authUser, null, authUser.getAuthorities());
    }

    protected String getMessage(String code, Object[] params) {
        try {
            return messageSource.getMessage(code, params, Locale.CHINA);
        } catch (NoSuchMessageException e) {
            return code;
        }
    }
}
