package com.chilunyc.fumao.common.aop.cache;

import com.chilunyc.fumao.common.aop.BaseAop;
import com.chilunyc.fumao.common.util.SpelHelper;
import lombok.extern.log4j.Log4j;
import org.apache.commons.codec.binary.Hex;
import org.apache.commons.lang.SerializationUtils;
import org.apache.commons.lang3.StringUtils;
import org.aspectj.lang.JoinPoint;
import org.aspectj.lang.ProceedingJoinPoint;
import org.aspectj.lang.annotation.Around;
import org.aspectj.lang.annotation.Aspect;
import org.aspectj.lang.annotation.Before;
import org.springframework.beans.factory.InitializingBean;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.redis.core.BoundValueOperations;
import org.springframework.data.redis.core.StringRedisTemplate;
import org.springframework.data.redis.core.script.DefaultRedisScript;
import org.springframework.stereotype.Component;

import java.io.Serializable;
import java.util.Collections;
import java.util.Map;

/**
 *
 */
@Log4j
@Component
@Aspect
public class CacheAop extends BaseAop implements InitializingBean {

    @Autowired
    private StringRedisTemplate redisTemplate;

    private DefaultRedisScript luaRemoveAllScript;

    /**
     * 缓存cache
     *
     * @param proceedingJoinPoint
     * @throws Throwable
     */
    @Around("@annotation(com.chilunyc.fumao.common.aop.cache.Cacheable)")
    Object cacheable(ProceedingJoinPoint proceedingJoinPoint) throws Throwable {
        try {
            Cacheable annotation = getAnnotation(proceedingJoinPoint, Cacheable.class);
            Map<String, Object> args = getArgs(proceedingJoinPoint);
            String redisKey = annotation.value() + SpelHelper.exec(annotation.key(), args);
            BoundValueOperations<String, String> value = redisTemplate.boundValueOps(redisKey);

            // 有值
            String result = value.get();
            if (StringUtils.isNotBlank(result)) {
                byte[] bytes = Hex.decodeHex(result.toCharArray());
                return SerializationUtils.deserialize(bytes);
            }
            // 没有值, 调用缓存
            else {
                Object proceed = proceedingJoinPoint.proceed(proceedingJoinPoint.getArgs());
                if (proceed == null) {
                    return null;
                }
                //判断是否要存
                if (StringUtils.isNotBlank(annotation.exclude())) {
                    args.put("result", proceed);
                    Boolean isNotSave = SpelHelper.exec(annotation.exclude(), args);
                    if (isNotSave) {
                        return proceed;
                    }
                }
                //开始存
                byte[] serialize = SerializationUtils.serialize((Serializable) proceed);
                String s = Hex.encodeHexString(serialize);
                if (annotation.timeout() != -1L) {
                    value.set(s, annotation.timeout(), annotation.unit());
                } else {
                    value.set(s);
                }

                // 放入一个list, 后面可以删除
                if (StringUtils.isNotBlank(annotation.value())) {
                    redisTemplate.boundListOps(annotation.value()).leftPush(redisKey);
                }
                return proceed;
            }
        } catch (Exception e) {
            log.error(e);
            return proceedingJoinPoint.proceed();
        }
    }

    /**
     * 清空cache
     *
     * @param joinPoint
     * @throws Throwable
     */
    @Before("@annotation(com.chilunyc.fumao.common.aop.cache.CacheEvict)")
    void cacheEvict(JoinPoint joinPoint) throws Throwable {
        CacheEvict annotation = getAnnotation(joinPoint, CacheEvict.class);
        // 清空所有
        if (annotation.allEntries()) {
            if (annotation.value().length > 0) {
                for (String v : annotation.value()) {
                    redisTemplate.execute(luaRemoveAllScript, Collections.singletonList(v));
                }
            }
        }
        // 清空单个
        else {
            String v = annotation.value()[0];
            Map<String, Object> args = getArgs(joinPoint);
            String redisKey = v + SpelHelper.exec(annotation.key(), args);
            redisTemplate.delete(redisKey);
            redisTemplate.boundListOps(v).remove(1, redisKey);
        }
    }

    @Override
    public void afterPropertiesSet() throws Exception {
        this.luaRemoveAllScript = new DefaultRedisScript<>();
        this.luaRemoveAllScript.setScriptText("local size = redis.call('llen', KEYS[1])\n" +
                "local vs = redis.call('lrange', KEYS[1], 0, size)\n" +
                "for x,y in ipairs(vs) do\n" +
                "        redis.call('del', y)\n" +
                "end\n" +
                "redis.call('ltrim', KEYS[1], size, -1)");
    }
}
