package com.chilunyc.fumao.common.util.jpa.query;

import lombok.AllArgsConstructor;
import lombok.Data;
import org.apache.commons.lang.StringUtils;
import org.springframework.data.jpa.domain.Specification;
import org.springframework.util.ReflectionUtils;

import javax.persistence.criteria.Predicate;
import java.lang.annotation.Annotation;
import java.lang.reflect.Field;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

/**
 * Created by Tkk on 2018/7/21.
 */
public class QueryBuilder {

    @Data
    @AllArgsConstructor
    private static class QueryDesc {
        Field field;
        QueryItem queryItem;
    }

    /**
     * 查询缓存
     */
    private static Map<Class, List<QueryDesc>> classQueryDescMap = new ConcurrentHashMap<>();


    /**
     * @param object
     * @param
     * @return
     */
    public static Specification build(Object object) {


        /**
         *
         */
        return (root, criteriaQuery, cb) -> {
            Class clazz = object.getClass();
            List<QueryDesc> descMap = getQueryDescMap(clazz);
            List<Predicate> predicates = new ArrayList<>(descMap.size());
            descMap.forEach(desc -> {
                Object value = getValue(desc.field, object);

                // 为null不加入查询
                if (value == null && !desc.queryItem.nullable()) {
                    return;
                }

                // 字符串为""不加入查询
                if (value != null && value instanceof String && !desc.queryItem.blankable() && StringUtils.isBlank(value.toString())) {
                    return;
                }

                //
                String path = desc.queryItem.column();
                if (StringUtils.isBlank(path)) {
                    path = desc.field.getName();
                }

                //
                switch (desc.queryItem.type()) {
                    case equal:
                        predicates.add(cb.equal(root.get(path), value));
                        break;
                    case like:
                        predicates.add(cb.like(root.get(path), "%" + value + "%"));
                        break;
                    case gt:
                        predicates.add(cb.gt(root.get(path), (Number) value));
                        break;
                    case lt:
                        predicates.add(cb.lt(root.get(path), (Number) value));
                        break;
                    case ge:
                        predicates.add(cb.ge(root.get(path), (Number) value));
                        break;
                    case le:
                        predicates.add(cb.le(root.get(path), (Number) value));
                        break;
                    case notEqual:
                        predicates.add(cb.notEqual(root.get(path), value));
                        break;
                    case notLike:
                        predicates.add(cb.notLike(root.get(path), "%" + value + "%"));
                        break;
                    case rightLike:
                        predicates.add(cb.like(root.get(path), value + "%"));
                        break;
                }
            });

            SortItems sortItems = (SortItems) clazz.getAnnotation(SortItems.class);
            if (sortItems != null && sortItems.value() != null) {
                for (SortItem sortItem : sortItems.value()) {
                    if (sortItem.desc()) {
                        criteriaQuery.orderBy(cb.desc(root.get(sortItem.column())));
                    } else {
                        criteriaQuery.orderBy(cb.asc(root.get(sortItem.column())));
                    }
                }
            }
            return cb.and(predicates.toArray(new Predicate[predicates.size()]));
        };
    }

    private static Object getValue(Field field, Object src) {
        try {
            return field.get(src);
        } catch (IllegalAccessException e) {
            return null;
        }
    }

    private static List<QueryDesc> getQueryDescMap(Class clazz) {
        List<QueryDesc> queryDescs = classQueryDescMap.get(clazz);
        if (queryDescs != null) {
            return queryDescs;
        }
        synchronized (clazz.getName()) {
            queryDescs = classQueryDescMap.get(clazz);
            if (queryDescs != null) {
                return queryDescs;
            }

            // 组装
            List<QueryDesc> real = new ArrayList<>();
            ReflectionUtils.doWithFields(clazz, (field) -> {
                QueryDesc queryDec = getQueryDec(field);
                if (queryDec == null) {
                    return;
                }
                field.setAccessible(true);
                real.add(queryDec);
            });

            //
            return real;
        }
    }

    private static QueryDesc getQueryDec(Field field) {
        QueryItem annotation = field.getAnnotation(QueryItem.class);
        if (annotation == null) {
            return null;
        }
        return new QueryDesc(field, annotation);
    }
}
